#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import time
import math

try:
    from twisted.internet import reactor
except ImportError:
    from twisted_emu import reactor

class BaseScheduler:
    """Basic scheduler class without next method."""
    def __init__(self, function, *args):
        self.function = function
        self.args = args
        self.running = None

    def start(self):
        """Start running the function given on __init__."""
        if not self.running:
            self.schedule()

    def stop(self):
        """Stop running the function given on __init__. The next call will be
        canceled."""
        if self.running:
            self.running.cancel()

    def run(self):
        """Internal method. Do not use."""
        self.running = None
        self.call()
        self.schedule()

    def next(self):
        """Return seconds to wait till next execution as float or int. In order
        to work this method has to be overwritten."""
        raise NotImplementedError

    def schedule(self):
        """Tell reactor about when to call me. Do not use."""
        self.running = reactor.callLater(self.next(), self.run)

    def call(self):
        """Call event."""
        return self.function(*self.args)

class CronParserError(ValueError):
    pass

def _parse_number(s, mapping={}, selector=min):
    """Internally used only. Parse a given number using optional given mapping
    and selection function. If given string is found in given mapping the
    selector (default: min) selects an element from that mapping, otherwise an
    int is returned. On error a CronParserError is raised."""
    s = s.lower()
    if s in mapping:
        return selector(mapping[s])
    try:
        return int(s)
    except ValueError:
        raise CronParserError, "not a number"

def _parse_range(field, start, end, mapping={}):
    l = field.split(',')
    r = []
    for e in l:
        if e.count('/') > 1 or e.count('-') > 1:
            raise CronParserError, "malformatted time spec"
        if '/' in e:
            e, s = e.split('/')
            s = _parse_number(s)
        else:
            s = 1
        if e == '*':
            f, t = start, end
        elif '-' in e:
            f, t = e.split('-')
            f = _parse_number(f, mapping, min)
            t = _parse_number(t, mapping, max)
        else:
            f = t = _parse_number(e, mapping, min)
        if not start <= f <= t <= end:
            raise CronParserError, "numbers out of range or not ordered"
        r.extend(range(f, t+1, s))
    r = dict.fromkeys(r).keys()
    r.sort()
    return r

class CronScheduler(BaseScheduler):
    """See man 5 crontab for the time_spec argument. Rest of arguments, see
    BaseScheduler."""
    def __init__(self, time_spec, function, *args):
        BaseScheduler.__init__(self, function, *args)
        try:
            n, h, d, m, w = time_spec.split()
        except ValueError:
            raise CronParserError, "5 fields are required"
        self.min = _parse_range(n, 0, 59)
        self.hour = _parse_range(h, 0, 23)
        self.mday = _parse_range(d, 1, 31)
        self.mon = _parse_range(m, 1, 12, dict(jan=(1,), feb=(2,), mar=(3,),
                    apr=(4,), may=(5,), jun=(6,), jul=(7,), aug=(8,), sep=(9,),
                    oct=(10,), nov=(11,), dec=(12,)))
        self.wday = _parse_range(w, 0, 7, dict(sun=(0,7), mon=(1,), tue=(2,),
                    wed=(3,), thu=(4,), fri=(5,), sat=(6,)))
        if 7 in self.wday:
            self.wday.remove(7)
            if not 0 in self.wday:
                self.wday.insert(0, 0)

    def compute_next(self, t=None):
        if t is None:
            t = time.time()
        t = math.ceil(t)
        tt = time.localtime(t)
        if tt.tm_sec:
            t += 60 - tt.tm_sec
            tt = time.localtime(t)
        while (tt.tm_mon not in self.mon or tt.tm_mday not in self.mday or
                    (tt.tm_wday+1)%7 not in self.wday):
             # skip to next daystart
             t += ((24 - tt.tm_hour) * 60 - tt.tm_min) * 60
             tt = time.localtime(t)
        while tt.tm_hour not in self.hour:
            # skip to next hour
            t += (60 - tt.tm_min) * 60
            tt = time.localtime(t)
            if not tt.tm_hour: # we changed a day
                return self.compute_next(t)
        while tt.tm_min not in self.min:
            # skip to next minute
            t += 60
            tt = time.localtime(t)
            if not tt.tm_min: # we changed an hour
                return self.compute_next(t)
        return t - time.time()
    def next(self):
        return self.compute_next()

class AdaptivePeriodicScheduler(BaseScheduler):
    """This scheduler tries to find the best interval on its own. The given
    function therefore has to return a boolean. A return value of True means to
    decrease the interval and a value of False means to decrease the interval.
    The meaning of the return value can be seen as "was this call necessary?".
    """
    def __init__(self, interval, function, *args):
        BaseScheduler.__init__(self, function, *args)
        self.interval = interval

    def next(self):
        return self.interval

    def call(self):
        if BaseScheduler.call(self):
            self.interval *= 0.9
        else:
            self.interval *= 1.1

class MinutelyScheduler(BaseScheduler):
    """Run function every minute. Aligned to local time."""
    def next(self):
        t = time.time()
        tt = time.localtime(math.ceil(t))
        return 60 - tt.tm_sec + (1 + (-t) % 1 )

class HourlyScheduler(BaseScheduler):
    """Run function every hour. Aligned to local time."""
    def next(self):
        t = time.time()
        tt = time.localtime(math.ceil(t))
        return (60 - tt.tm_min) * 60 - tt.tm_sec + (1 + (-t) % 1 )

class DailyScheduler(BaseScheduler):
    """Run function every day. Aligned to local time."""
    def next(self):
        t = time.time()
        tt = time.localtime(math.ceil(t))
        return ((24 - tt.tm_hour)*60 - tt.tm_min)*60 - tt.tm_sec + (1+(-t)%1)

if __name__ == '__main__':
    def test(s):
        print "string:", s
        t = CronScheduler(s, None)
        print "now:   ", time.time(), time.asctime(time.localtime())
        n = time.time() + t.next()
        print "next:  ", n, time.asctime(time.localtime(n))

    test("14,*/10,44 * * * *")
    test("30 * * * sun")
    test("30 8-20/4 * * fri-sun")
    m = DailyScheduler(1)
    print m.next()
    print time.time() + m.next()
    print time.localtime(time.time() + m.next())
    print time.asctime(time.localtime(time.time() + m.next()))
