#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

try:
    #raise ImportError
    from twisted.internet.protocol import ClientFactory, Protocol
    from twisted.internet import reactor
except ImportError:
    from twisted_emu import ClientFactory, Protocol, reactor

class pluginFolder:
    """Plugin folder class."""

    def __init__(self):
        self._plugins = {}

    def __setitem__(self, key, value):
        """Add plugins or folders.
        Use folder[key] = None to add a folder and folder[key]=plugin to add
        plugins."""
        if '.' in key:
            k,key = key.split(".", 1)
            if not self._plugins.has_key(k):
                self[k] = self.__class__()
            elif not isinstance(self._plugins[k], self.__class__):
                raise ValueError("adding value to non-folder")
            self._plugins[k][key] = value
        else:
            if self._plugins.has_key(key):
                raise ValueError("key already exsist")
            self._plugins[key] = value

    def __getitem__(self, key):
        """Get plugin by key. (without eval)"""
        if '.' in key:
            k,key = key.split(".", 1)
            return self._plugins[k][key]
        return self._plugins[key]

    def __delitem__(self, key):
        """Delete plugin by key. (without eval)"""
        if '.' in key:
            k,key = key.split(".", 1)
            del self._plugins[k][key]
        else:
            del self._plugins[key]

    def __getattr__(self, key):
        """Make keys look like attributes."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError("%s instance has no attribute %s" %
                (self.__class__.__name__, repr(key)))

class plugin:
    """Plugin meta class containing information for managing plugins."""

    def __init__(self, name):
        self.name = name
        module_name = "plugins." + name
        last_name = name.split(".")[-1]
        self.code = __import__(module_name, {}, {}, (last_name,))
        self.instance = None
        # Plugin might be removed and loaded again, because it has no state
        # preserving methods. Therefore force a reload.
        reload(self.code)
        self.setup()

    def setup(self):
        """Prepare creating the plugin instance by setting dependencies and
        conflicts. This is called from __init__ and reload."""
        self.depends = self.code.depends
        self.conflicts = self.code.conflicts
        self.unload()

    def load(self, base):
        """Actually create an instance of the plugin."""
        self.instance = self.code.plugin(base)

    def reload(self, base):
        """API: unstable
        Tries to reload given this plugin by doing the following steps.
        * Check if that plugin has __getstate__ and __setstate__ methods to be
            able to be reloaded. Return "nogetstate" unless. This will change.
        * Get the state by calling the plugins __getstate__ method.
        * Reload the python code.
        * Now check the dependencies (may be changed) and return "depfailed"
            if they could not be met.
        * Call self.setup, create a new instance and call the __setstate__
           method with the return value of __getstate__ as single argument.
        * Return True.
        The return values will be changed to exceptions when I think that it
        won't break existing code."""
        if not (hasattr(self.instance, "__getstate__") and
                hasattr(self.instance, "__setstate__")):
            return "nogetstate" # DEBUG
        state = self.instance.__getstate__()
        reload(self.code)
        if not base.checkdepends(self.code.depends, self.code.conflicts):
            return "depfailed" # DEBUG
        self.setup()
        self.load(base)
        self.instance.__setstate__(state)
        return True

    def unload(self):
        """Remove plugin instance in order to remove this plugin."""
        if hasattr(self.instance, "delete"): # this is False for None
            self.instance.delete()
        self.instance = None

class mibcipProtocol(Protocol):
    """This is the protocol class for mibcip."""

    def connectionMade(self):
        """Catch the twisted event and call the method on_connect of the
        factory. on_connect should be used by one plugin."""
        self.factory.protoinstance = self
        if not self.factory.on_connect:
            raise NotImplementedError
        self.factory.on_connect(())

    def connectionLost(self, reason):
        """Catch the twisted event and call the method on_disconnect of the
        factory. on_disconnect should be used by one plugin."""
        self.factory.protoinstance = None
        if not self.factory.on_disconnect:
            raise NotImplementedError
        self.factory.on_disconnect(())

    def dataReceived(self, data):
        """Catch the twisted event and call the method on_data of the factory.
        on_data should be used by one plugin."""
        if not self.factory.on_data:
            raise NotImplementedError
        self.factory.on_data((data,))


class mibcipFactory(ClientFactory):
    """Factory class for mibcip.
    To access plugins use factory.p.name.of.the.plugin. Exactly one core plugin
    must set factory.on_connect, factory.on_data and factory.on_disconnect."""
    protocol = mibcipProtocol

    def __init__(self, configfile):
        """Create a mibcipFactory given the name of its config file."""
        self.configfile = configfile

        self.p = pluginFolder()
        self.plugins = {}

        self.protoinstance = None
        self.on_connect = None
        self.on_data = None
        self.on_disconnect = None
        self.on_confailed = None

        self.insmod("core.startup")

    def haveAnyPlugin(self, *pl):
        """Check if any of the plugins given as arguments are loaded. No data
        structures are modified and the return value will be either True or
        False."""
        for p in pl:
            if p in self.plugins:
                return True
        return False

    def checkdepends(self, depends, conflicts):
        """Check if depends and conflicts are met. Depends should be a list of
        lists which. For each list at least one item has to be loaded.
        Conflicts should be a list of plugins which must not be loaded. The
        return value is True in case dependencies and conflicts are met. No data
        structures are modified."""
        if self.haveAnyPlugin(*conflicts):
            return False
        for d in depends:
            if not self.haveAnyPlugin(*d):
                return False
        return True

    def reload(self, name):
        """Reload given plugin."""
        ret = self.plugins[name].reload(self)
        del self.p[name]
        self.p[name] = self.plugins[name].instance
        return ret

    def insmod(self, name):
        """Insert a plugin. Some other plugins may be loaded as dependencies."""
        for p in self.plugins.values():
            if name in p.conflicts:
                return False
        self.plugins[name] = p = plugin(name)
        # Checking for conflicting plugins.
        if self.haveAnyPlugin(*p.conflicts):
            del self.plugins[name]
            return False

        # Checking for dependencies.
        for l in p.depends:
            # If dependency is not fulfilled,
            if not self.haveAnyPlugin(*l):
                loaded = False
                # try loading plugins in order.
                for n in l:
                    try:
                        if self.insmod(n):
                            loaded = True
                            break
                    except:
                        del self.plugins[name]
                        raise
                # Fail if none could be loaded.
                if not loaded:
                    del self.plugins[name]
                    return False
        try:
            p.load(self)
        except:
            del self.plugins[name]
            raise
        self.p[name] = p.instance
        return True

    def rmmod(self, name):
        """Remove a plugin. If this would break dependencies, fail."""
        # An inexsistent plugin cannot be removed.
        if not self.plugins.has_key(name):
            return False
        # For each plugin check
        for plgn in self.plugins.values():
            # for each item in dependency list
            for deplist in plgn.depends:
                # if this plugin may be required
                if name in deplist:
                    t = deplist[:]
                    # and if dependencies are still fulfilled without this one.
                    t.remove(name)
                    loaded = False
                    for deplgn in t:
                        if self.plugins.has_key[deplgn]:
                            loaded = True
                            break
                    # If not, bail out.
                    if not loaded:
                        return False
        try:
            self.plugins[name].unload()
        finally:
            del self.p[name]
            del self.plugins[name]
        return True

    def clientConnectionFailed(self, connector, reason):
        """Dummy function needed by twisted."""
        if self.on_confailed is not None:
            self.on_confailed(())
        print 'connection failed:', reason.getErrorMessage()

    def clientConnectionLost(self, connector, reason):
        """Dummy function needed by twisted."""
        print 'connection lost:', reason.getErrorMessage()

if __name__ == "__main__":
    botfactory = mibcipFactory("mibcip.conf")
    server = botfactory.p.core.config.get("core", "server")
    port = botfactory.p.core.config.getint("core", "port")

    print "connecting to %s:%d..." % (server, port)
    reactor.connectTCP(server, port, botfactory)
    reactor.run()
