#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

conflicts = []
depends = [ ["core.config"], ["core.event"], ["core.parser"], ["core.info"] ]

class authUser:
    """This data structure is internally used to represent the ircpasswd."""
    def __init__(self, name):
        self.name = name
        self.flags = ()
        self.level = 0
        self.methods = []

    def fromline(klass, line):
        """Parse a line into a authUser."""
        name, methods, rights = line.strip().split(":")
        self = klass(name)
        self.set_rights(rights)
        self.set_methods(methods)
        return self
    fromline = classmethod(fromline)

    def toline(self):
        """Return a string which can be parsed using fromline."""
        methods = [[";".join(m) for m in a] for a in self.methods]
        methods = ["&".join(a) for a in methods]
        methods = "|".join(methods)
        flags = [str(self.level)] + self.flags
        flags = ";".join(flags)
        return "%s:%s:%s" % (self.name, methods, flags)

    def set_rights(self, rights):
        """Set the rights of a user given the third field of an entry."""
        self.flags = rights.split(";")
        self.level = int(self.flags.pop(0))

    def set_methods(self, methods):
        """Set the methods to authenticated of a user given the second field of
        an entry."""
        def make_method(token):
            t = token.split(";", 1)
            if len(t) < 2:
                return t[0], ""
            return tuple(t)

        def make_and(token):
            return [make_method(m) for m in token.split("&")]

        def make_or(token):
            return [make_and(m) for m in token.split("|")]

        self.methods = make_or(methods)

    def collect_methods(self):
        """Return all names of methods that may be used to authenticated this
        user."""
        methods = reduce(lambda a,b: a+b, self.methods, [])
        methods = dict.fromkeys(methods).keys()
        return methods

    def check_methods(self, methods):
        """Check the authentication state of this user given a methods dict. The
        dict should contain method names as keys and True or False as values.
        If the user can be authenticated with these methods True is returned. If
        the user has no chance authenticating given to many failed methods False
        is returned. Returns None if the user is not authenticated but still has
        methods that may do so."""
        mapping = {False: 0, None: 1, True: 2}
        t = [[methods.get(m, None) for m in a] for a in self.methods]
        t = [[mapping[m] for m in a] for a in t]
        # To understand this replace max with or and min with and. The and
        # operation on the operator list gives False if there is a False among
        # the operators and True if it consists of True, else None. The or
        # operation on the operator list gives False if it consists of False
        # and True if there is a True among the operators, else None.
        t = max([min(l) for l in t])
        return [False, None, True][t]

class plugin:
    def __init__(self, base):
        self.base = base
        # map usernames to authentication structures
        self.users = {}
        # map nicks that are logged in to usernames
        self.nicks = {}
        # map method names to functions
        self.methods = {}
        # map nicks to their method cache (map method (name, arg) tuples to
        # their results)
        self.cache = {}
        # map nicks that try to log in to usernames
        self.tries = {}

        event = base.p.core.event

        event.register_plugin("core.auth")

        attach_event = event.attach_event
        attach_event("core.info", "core.auth", "NICKLOST", self.on_nicklost)
        attach_event("core.parser", "core.auth", "NICK", self.on_nick)
        attach_event("core.event", "core.auth", "disconnect",
                     self.on_disconnect)

        filename = base.p.core.config.get("core.auth", "ircpasswd")
        self.load(filename)

    def __getstate__(self):
        users = [u.toline() for u in self.users.values()]
        return users, self.nicks, self.methods, self.cache, self.tries

    def __setstate__(self, (users, nicks, methods, cache, tries)):
        self.users = {}
        for u in users:
            t = authUser.fromline(u)
            self.users[t.name] = t
        self.nicks = nicks
        self.methods = methods
        self.cache = cache
        self.tries = tries

    def save(self, filename):
        """Save user representation to file given a filename."""
        f = file(filename, "w")
        for u in self.users.values():
            f.write("%s\n" % u.toline())
        f.close()

    def load(self, filename):
        """Load user representation from a file given a filename."""
        f = file(filename)
        self.users = {}
        for l in f.readlines():
            l = l.strip("\n")
            u = authUser.fromline(l)
            self.users[u.name] = u
        f.close()

    def login(self, nick, user):
        """Try to login given nick as given user. Optionally call report
        with True or False."""
        if user not in self.users:
            return
        if nick in self.nicks:
            self.logout(nick)
        self.tries[nick] = user
        if nick not in self.cache:
            self.cache[nick] = {}
        for method in self.users[user].collect_methods():
             self.call_method(method[0], method[1], nick)
        self.check_login(nick)

    def check_login(self, nick):
        """Try to login given nick. The login method must be called first. If
        this is given methods are checkedand the login event is called on
        success."""
        if nick not in self.tries:
            return
        user = self.tries[nick]
        if user not in self.users:
            del self.tries[nick]
            return
        if self.users[user].check_methods(self.cache[nick]):
            del self.tries[nick]
            self.nicks[nick] = user
            self.base.p.core.event.call_event("core.auth", "LOGIN",
                                              (nick, user))

    def logout(self, nick):
        """Logout given nick if he was logged in."""
        try:
            del self.nicks[nick]
        except KeyError:
            pass
        else:
            self.base.p.core.event.call_event("core.auth", "LOGOUT", (nick,))

    def logged_in(self, nick):
        """Check if given nick is logged in."""
        return nick in self.nicks

    def auth_level(self, nick):
        """Return level of nick. Return 0 if nick is not logged in."""
        if not self.logged_in(nick):
            return 0
        return self.users[self.nicks[nick]].level

    def auth_flags(self, nick):
        """Return flags of a nick as a tuple. Return an empty tuple if nick is
        not logged in."""
        if not self.logged_in(nick):
            return ()
        return self.users[self.nicks[nick]].flags

    def register_method(self, name, function):
        """Register an authentication method called name with given function.
        Every time the method is needed the function is called with
        function(<nickname>, <argument from passwd file>) and should either
        call method_success or method_fail."""
        self.methods[name] = function

    def unregister_method(self, name):
        """Unregister an authentication method called name and make it fail at
        any time it is called."""
        try:
            del self.methods[name]
        except KeyError:
            pass

    def call_method(self, name, arg, nick):
        """Call method registered as name with given nick and arg. If method is
        not registered assume a value of False."""
        if name in self.methods:
            self.methods[name](nick, arg)
        else:
            self.cache[nick][name,arg] = False

    def report(self, name, arg, nick, value, overwrite=True):
        """Report the return value of an authentication method with given
        argument and nickname as value."""
        if nick not in self.cache:
            self.cache[nick] = {}
        if overwrite or (name,arg) not in self.cache[nick]:
            self.cache[nick][name,arg] = value
        self.check_login(nick)

    def report_failure(self, name, arg, nick, overwrite=True):
        """Report an authentification failure for given method name, argument
        and nickname."""
        self.report(name, arg, nick, False, overwrite)

    def report_success(self, name, arg, nick, overwrite=True):
        """Report an authentification success for given method name, argument
        and nickname."""
        self.report(name, arg, nick, True, overwrite)

    def remove_nick(self, nick):
        """Actually logout and remove nick from all the data structures."""
        self.logout(nick)
        try:
            del self.tries[nick]
        except KeyError:
            pass
        try:
            del self.cache[nick]
        except KeyError:
            pass

    def on_nicklost(self, (nick,)):
        """Process quit event for automatic logouts."""
        self.remove_nick(nick)

    def on_nick(self, (nick, new)):
        """Process nick event for tracking users."""
        if nick in self.nicks:
            self.nicks[new] = self.nicks[nick]
            del self.nicks[nick]
        if nick in self.cache:
            self.cache[new] = self.cache[nick]
            del self.cache[nick]
        if nick in self.tries:
            self.tries[new] = self.tries[nick]
            del self.tries[nick]

    def on_disconnect(self, arg):
        """Remove all data concerning current session."""
        self.nicks = {}
        self.cache = {}
        self.tries = {}
