#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

conflicts = []
depends = []

class eventContainer:
    """Each plugin has an eventContainer that maps eventnames to a dict that
    maps plugin names to functions."""
    def __init__(self):
        self.events = {}

    def __getstate__(self):
        return self.events

    def __setstate__(self, state):
        self.events = state

    def attach_event(self, name, event, func):
        """Add the function func to this eventContainer for the event event as
        the plugin named name."""
        if not self.events.has_key(event):
            self.events[event] = {}
        self.events[event][name] = func

    def unattach_event(self, name, event):
        """Remove the function attached to this eventContainer for the event
        event as the plugin named name."""
        try:
            del self.events[event][name]
            if not self.events[event]:
                del self.events[event]
        except KeyError:
            pass

    def call_event(self, event, arg):
        """Call all functions with argument arg that match the given event
        name."""
        if self.events.has_key(event):
            for e in self.events[event].values():
                e(arg)

    def unattach_all_events(self, name):
        """Remove all event attachments by the plugin named name from this
        eventContainer."""
        for e in self.events.values():
            try:
                del e[name]
            except KeyError:
                pass

class plugin:
    """Provide an event handler. API: stable."""
    def __init__(self, base):
        self.plugins = {}
        self.queue = []
        # If we are currently processing an event.
        self.working = False

        # Proxying connectionMade and dataReceived.
        base.on_connect = self.on_connect
        base.on_data = self.on_data
        base.on_disconnect = self.on_disconnect
        base.on_confailed = self.on_confailed
        self.register_plugin("core.event")

    def __getstate__(self):
        state = {}
        for k, v in self.plugins.items():
             state[k] = v.__getstate__()
        return self.queue, state

    def __setstate__(self, (queue, state)):
        self.queue = queue
        self.plugins = {}
        for k, v in state.items():
            self.plugins[k] = eventContainer()
            self.plugins[k].__setstate__(v)

    def on_connect(self, arg):
        """This method maps a connection established event to an event
        "connect" owned by "core.event"."""
        self.call_event("core.event", "connect", arg)

    def on_data(self, arg):
        """This method maps a data received even to an event "data" owned by
        "core.event"."""
        self.call_event("core.event", "data", arg)

    def on_disconnect(self, arg):
        """This method maps a disconnect event to an event "disconnect" owned
        by "core.event"."""
        self.call_event("core.event", "disconnect", arg)

    def on_confailed(self, arg):
        """This method maps a connection failure event to an event "confailed"
        owned by "core.event"."""
        self.call_event("core.event", "confailed", arg)

    def register_plugin(self, name):
        """Register a plugin for attaching and registering events."""
        if name not in self.plugins:
            self.plugins[name] = eventContainer()

    def unregister_plugin(self, name):
        """Remove a plugin from all attached events and delete all registered
        events."""
        try:
            del self.plugins[name]
        except KeyError:
            pass
        for p in self.plugins.values():
            p.unattach_all_events(name)

    def attach_event(self, name, myname, event, func):
        """Attach event "event" of plugin "name" as plugin "myname"."""
        if name in self.plugins:
            self.plugins[name].attach_event(myname, event, func)

    def unattach_event(self, name, myname, event):
        """Unattach an event of another plugin."""
        if name in self.plugins:
            self.plugins[name].unattach_event(myname, event)

    def really_call_event(self, name, event, arg):
        """Really call an event without delaying it. Don't call me."""
        if name not in ("core.event","core.rawparser"):
            print "EVENT:", name, event, arg
        if name in self.plugins:
            self.plugins[name].call_event(event, arg)

    def call_event(self, name, event, arg):
        """Call an event event of plugin name with argument arg."""

        self.queue.append((name, event, arg))

        if self.working:
            return

        self.working = True

        while self.queue:
            self.really_call_event(*self.queue.pop(0))

        self.working = False
