#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import time

conflicts = []
depends = [ ["core.event"], ["core.parser"], ["core.send"] ]

def compare_ban_part(ban, name):
    """Check if a ban containing only '?' or characters, but not any of '*!@'
    matches the beginning of a name."""
    for x,y in zip(ban, name):
        if x is None:
            return True
        if x not in ('?', y):
            return False
    return True

def compare_ban_token(ban, name):
    """Check if a ban containing only '*?' or characters, but not any of '!@'
    matches a complete name."""
    if ban == '*':
        return True
    ban = ban.split('*')
    if not compare_ban_part(ban[0], name):
        return False
    rest = name[len(ban.pop(0)):]
    while len(ban) > 1:
        found = False
        for x in range(len(rest)):
            if compare_ban_part(ban[0], rest[x:]):
                rest = rest[len(ban.pop(0)) + x:]
                found = True
                break
        if not found:
            return False
    if not rest:
        return not (ban and ban[0])
    if ban and compare_ban_part(ban[0], rest[-len(ban[0]):]):
        return True
    return False

def isbanned(ban, nick, user=None, host=None):
    """Check if a complete ban matches a given user like 'nick!user@host' or
    splitted up into its three parts."""
    if user is None and host is None:
        nick, user = nick.split('!', 1)
        user, host = user.split('@', 1)
    bn, bu = ban.split('!', 1)
    bu, bh = bu.split('@', 1)
    return (compare_ban_token(bn, nick) and
            compare_ban_token(bu, user) and
            compare_ban_token(bh, host))

class myInf:
    """This structure conatins information about the bot's nick."""
    def __init__(self, base):
        self.base = base
        self.nick = ""
        self.umodes = ""

    def __getstate__(self):
        return (self.nick, self.umodes)

    def __setstate__(self, state):
        self.nick, self.umodes = state

    def addMode(self, mode):
        """Add umodes in mode to me. Don't call."""
        umods = self.base.p.core.servcapab.umods
        for x in mode:
            if x in umods and not x in self.umodes:
                self.umodes+=x

    def delMode(self, mode):
        """Remove umodes in mode from me. Don't call."""
        for x in mode:
            self.umodes = self.umodes.replace(x, "")

    def __str__(self):
        """Return my nick."""
        return self.nick
    def __repr__(self):
        """Return my nick."""
        return "%s(+%s)" % (self.nick, self.umodes)

class userInf:
    """This structure contains information about a single user on irc."""
    def __init__(self, nick, user=None, host=None, real=None):
        self.nick = nick
        self.user = user
        self.host = host
        self.real = real

    def __getinitargs__(self):
        return (self.nick, self.user, self.host, self.real)

    def isbanned(self, ban):
        """Check if user i matched by given ban mask."""
        if self.user is None or self.host is None:
            return None
        return isbanned(ban, self.nick, self.user, self.host)

    def __str__(self):
        """Return a string which looks like "nick!user@host"."""
        ret = [self.nick]
        if self.user:
            ret.append('!' + self.user)
        if self.host:
            ret.append('@' + self.host)
        return repr("".join(ret))

    def __repr__(self):
        """Return a string which looks like "nick!user@host:realname"."""
        ret = [self.nick]
        if self.user:
            ret.append('!' + self.user)
        if self.host:
            ret.append('@' + self.host)
        if self.real:
            ret.append(':' + self.real)
        return repr("".join(ret))

class chanInf:
    """This structure contains information collected about a single channel."""
    def __init__(self, base, chan):
        self.base = base
        self.chan = chan
        self.users = {}
        self.topic = None
        self.tpcwho = None
        self.tpcwhen = None
        self.modes = {}
        self.bans = {}

    def __getstate__(self):
        return (self.chan, self.users, self.topic, self.tpcwho,
                self.tpcwhen, self.modes, self.bans)

    def __setstate__(self, state):
        (self.chan, self.users, self.topic, self.tpcwho,
         self.tpcwhen, self.modes, self.bans            ) = state

    def __repr__(self):
        tokens = (self.chan, self.topic, self.tpcwho, self.tpcwhen,
                  self.modes, self.bans, self.users)
        tokens = map(repr, tokens)
        return "<Chan %s>" % ":".join(tokens)

    def addBan(self, ban, who, when=None):
        """Add given ban mask added by who on when or now if when is None to
        this channel."""
        if when is None:
            when = time.time()
        users = self.base.p.core.info.users
        if '!' not in who and who in users:
            who = str(users[who])
        self.bans[ban] = (who, when)

    def delBan(self, ban):
        """Remove whatever happens this ban from this channel."""
        try:
            del self.bans[ban]
        except KeyError:
            pass

    def isbanned(self, nick):
        """Check if given nick is banned on this channel."""
        users = self.base.p.core.info.users
        user = users[nick].user
        host = users[nick].host
        for ban in self.bans:
            if isbanned(ban, nick, user, host):
                return True
        return False

    def addMode(self, mode, arg=None):
        """Add given mode with optional argument."""
        self.modes[mode] = arg

    def delMode(self, mode):
        """Remove a mode from this channel."""
        try:
            del self.modes[mode]
        except KeyError:
            pass

    def addNick(self, nick, prefix=""):
        """Add a nick with optional prefix to this channel."""
        self.users[nick] = prefix

    def delNick(self, nick):
        """Remove a nick from this channel."""
        try:
            del self.users[nick]
        except KeyError:
            pass

    def hasNick(self, nick):
        """Check if given nick is located in this channel."""
        return nick in self.users

    def nicks(self, predicate=lambda _: True):
        """Return the list of nicks matching given prefix filter function or
        all if omited."""
        return [n for (n, p) in self.users.items() if predicate(p)]

    def ops(self):
        """Return a list of ops."""
        return self.nicks(lambda p: '@' in p)

    def addUMode(self, mode, nick):
        """Add given mode to nick on this channel."""
        servcapab = self.base.p.core.servcapab
        prefix = servcapab.cumodeprefix(mode)
        oldprefixes = self.users.get(nick, '')
        self.users[nick] = servcapab.insertprefix(oldprefixes, prefix)

    def delUMode(self, mode, nick):
        """Remove given mode from nick on this channel."""
        servcapab = self.base.p.core.servcapab
        prefix = servcapab.cumodeprefix(mode)
        oldprefixes = self.users.get(nick, '')
        self.users[nick] = servcapab.removeprefix(oldprefixes, prefix)

    def changeNick(self, oldnick, newnick):
        """Rename oldnick to newnick."""
        if oldnick in self.users:
            self.users[newnick] = self.users[oldnick]
            del self.users[oldnick]

class plugin:
    events = "addban addwho amode cumode join kick names nick part quit"
    events += " setban setmodes setnick setpcsrc settime settopic smode topic"
    events += " umode unban"
    events = events.split()
    def __init__(self, base):
        self.base = base
        self.myinf = myInf(base)
        self.users = {}
        self.chans = {}
        self.timedelta = 0.0
        event = base.p.core.event
        event.register_plugin("core.info")
        attach_event = event.attach_event
        for e in self.events:
            attach_event("core.parser", "core.info", e.upper(),
                         getattr(self, "received_" + e.lower()))
        attach_event("core.event", "core.info", "disconnect",
                     self.received_disconnect)

    def __getstate__(self):
        myinf = self.myinf.__getstate__()
        users = {}
        for k, v in self.users.items():
            users[k] = v.__getinitargs__()
        chans = {}
        for k, v in self.chans.items():
            chans[k] = v.__getstate__()
        return self.timedelta, myinf, users, chans

    def __setstate__(self, state):
        self.timedelta = state[0]
        self.myinf.__setstate__(state[1])
        self.users = {}
        for k, v in state[2].items():
            self.users[k] = userInf(*v)
        self.chans = {}
        for k, v in state[3].items():
            self.chans[k] = chanInf(self.base, None)
            self.chans[k].__setstate__(v)

    def server_now(self):
        return time.time() - self.timedelta

    def remove_nick(self, nick):
        try:
            del self.users[nick]
        except KeyError:
            pass
        else:
            self.base.p.core.event.call_event("core.info", "NICKLOST", (nick,))

    def received_disconnect(self, arg):
        """Reset everything."""
        self.myinf = myInf(self.base)
        self.users = {}
        self.chans = {}
        self.timedelta = 0.0

    def received_addban(self, (nick, chan, mask)):
        if chan in self.chans:
            self.chans[chan].addBan(mask, nick)

    def received_addwho(self, (nick, user, host, realname)):
        if nick in self.users:
            user_info = self.users[nick]
            user_info.user = user
            user_info.host = host
            user_info.real = realname

    def received_amode(self, (nick, chan, mode, arg)):
        if chan in self.chans:
            if mode[0] == '+':
                self.chans[chan].addMode(mode[1], arg)
            else:
                self.chans[chan].delMode(mode[1])

    def received_cumode(self, (nick, chan, mode, target)):
        if chan in self.chans:
            if mode[0] == '+':
                 self.chans[chan].addUMode(mode[1], target)
            else:
                 self.chans[chan].delUMode(mode[1], target)

    def received_join(self, arg):
        quote = self.base.p.core.send.quote
        IMP = self.base.p.core.send.IMP
        nick = arg.nick
        if nick == self.myinf.nick:
            self.chans[arg.chan] = chanInf(self.base, arg.chan)
            quote("WHO %s" % arg.chan, IMP)
            quote("MODE %s" % arg.chan, IMP)
            quote("MODE %s +b" % arg.chan, IMP)
        else:
            self.users[nick] = userInf(nick, arg.raw.user, arg.raw.host)
            if arg.chan in self.chans:
                self.chans[arg.chan].addNick(nick)
            quote("WHO %s" % nick, IMP)

    def received_kick(self, (nick, chan, target, reason)):
        # A kick has the same effect as a part for the information structures.
        self.received_part((target, chan, None))

    def received_names(self, (chan, names)):
        if not chan in self.chans:
            return
        channel = self.chans[chan]
        for (nick, prefix) in names.items():
            channel.addNick(nick, prefix)
            self.users[nick] = userInf(nick)

    def received_nick(self, (oldnick, newnick)):
        if oldnick == self.myinf.nick:
            self.myinf.nick = newnick

        if oldnick in self.users:
            self.users[newnick] = self.users[oldnick]
            del self.users[oldnick]
            self.users[newnick].nick = newnick

        for chan in self.chans.values():
            chan.changeNick(oldnick, newnick)

    def received_part(self, (nick, chan, reason)):
        if nick == self.myinf.nick:
            try:
                del self.chans[chan]
            except KeyError:
                pass
            for n in self.users.keys():
                found = False
                for c in self.chans.values():
                    if c.hasNick(n):
                        found = True
                        break
                if not found:
                    self.remove_nick(n)
        else:
            if chan in self.chans:
                self.chans[chan].delNick(nick)
            for c in self.chans.values():
                if c.hasNick(nick):
                    # Do not delete nick, since he is in anther channel.
                    return
            self.remove_nick(nick)

    def received_quit(self, (nick, reason)):
        self.remove_nick(nick)
        for chan in self.chans.values():
            chan.delNick(nick)

    def received_setban(self, (chan, mask, srcmask, when)):
        if chan in self.chans:
            self.chans[chan].addBan(mask, srcmask, when + self.timedelta)

    def received_setmodes(self, (chan, modes)):
        if chan in self.chans:
            channel = self.chans[chan]
            for mode in channel.modes.keys():
                channel.delMode(mode)
            for mode, arg in modes.items():
                channel.addMode(mode, arg)

    def received_setnick(self, (nick,)):
        self.myinf.nick = nick

    def received_setpcsrc(self, (chan, who, when)):
        if chan in self.chans:
            self.chans[chan].tpcwho = who
            self.chans[chan].tpcwhen = float(when) + self.timedelta

    def received_settime(self, (tm,)):
        self.timedelta = time.time() - tm
        print "DEBUG: settimg timedelta", self.timedelta # DEBUG

    def received_settopic(self, (chan, topic)):
        if chan in self.chans:
            self.chans[chan].topic = topic

    def received_smode(self, (nick, chan, mode)):
        if chan in self.chans:
            if mode[0] == '+':
                self.chans[chan].addMode(mode[1])
            else:
                self.chans[chan].delMode(mode[1])

    def received_topic(self, (nick, chan, topic)):
        if chan in self.chans:
            channel = self.chans[chan]
            channel.topic = topic
            channel.tpcwho = nick
            channel.tpcwhen = time.time()

    def received_umode(self, (mode,)):
        if mode[0] == '+':
            self.myinf.addMode(mode[1])
        else:
            self.myinf.delMode(mode[1])

    def received_unban(self, (nick, chan, mask)):
        if chan in self.chans:
            self.chans[chan].delBan(mask)

    def isself(self, nick):
        """Check whether given nick is my nick."""
        return nick == self.myinf.nick
