#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

conflicts = []
depends = [ ["core.event"], ["core.rawparser"], ["core.servcapab"] ]

import time

__dt_formats = ('%A %B %d %Y -- %H:%M:%S +00:00',)

def parse_timestamp(ts):
    """Try to parse a date string given by RPL_TIME or 391 into seconds
    since epoch."""
    for fmt in __dt_formats:
        try:
            tp = time.strptime(ts, fmt)
            return time.mktime(tp)
        except ValueError:
            pass
    return None


class ParserEvent:
    """This parser event behaves like a tuple with attributes."""
    def __init__(self, *args, **kwargs):
        """*args should be a tuple of (key, value) pairs. Don't use **kwargs,
        if you want to preserve order and want values to be shown up in the
        tuple representation."""
        self.__d = kwargs.copy()
        self.__l = []
        for (k, v) in args:
            self.__d[k] = v
            self.__l.append(v)

    def __getattr__(self, name):
        try:
            return self.__d[name]
        except KeyError:
            raise AttributeError

    def __getitem__(self, n):
        return self.__l[n]

    def __iter__(self):
        return iter(self.__l)

    def __repr__(self):
        return "<ParserEvent {%s}>" % ", ".join(map(repr, self.__l))

class plugin:
    commands = "001 324 332 333 352 353 367 391 invite join kick mode nick"
    commands += " notice part privmsg quit topic"
    commands = commands.split()
    def __init__(self, base):
        self.base = base
        event_handler = self.base.p.core.event
        event_handler.register_plugin("core.parser")
        for cmd in self.commands:
            event_name = cmd.upper()
            method = getattr(self, "received_" + cmd.lower())
            event_handler.attach_event("core.rawparser", "core.parser",
                                       event_name, method)

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        pass

    def emit_signal(self, name, *args, **kwargs):
        argument = ParserEvent(*args, **kwargs)
        self.base.p.core.event.call_event("core.parser", name, argument)

    def received_001(self, (arg,)):
        """Process RPL_WELCOME and emit SETNICK."""
        self.emit_signal("SETNICK", ("nick", arg.argv[0]), raw=arg)

    def received_324(self, (arg,)):
        """Process RPL_CHANNELMODEIS and emit SETMODES."""
        # first argument is my nickname
        args = arg.argv[1:]
        # followed by channel and mode string
        chan = args.pop(0)
        modes = args.pop(0)

        dct = {}
        servcapab = self.base.p.core.servcapab

        for c in modes:
            if c in servcapab.smods:
                dct[c] = None
            elif c in servcapab.amods:
                dct[c] = args.pop(0)

        self.emit_signal("SETMODES", ("chan", chan), ("dct", dct), raw=arg)

    def received_332(self, (arg,)):
        """Process RPL_TOPIC and emit SETTOPIC."""
        chan = ("chan", arg.argv[1])
        topic = ("topic", arg.text)
        self.emit_signal("SETTOPIC", chan, topic, raw=arg)

    def received_333(self, (arg,)):
        """Process unknown RPL_333 telling us who set the topic when and emit
        SETPCSRC."""
        chan = ("chan", arg.argv[1])
        who = ("who", arg.argv[2])
        when = ("when", int(arg.argv[3]))
        self.emit_signal("SETPCSRC", chan, who, when, raw=arg)

    def received_352(self, (arg,)):
        """Process RPL_WHOREPLY and emit ADDWHO."""
        nick = ("nick", arg.argv[5])
        user = ("user", arg.argv[2])
        host = ("host", arg.argv[3])
        realname = ("realname", arg.text.split(' ',1)[1])
        self.emit_signal("ADDWHO", nick, user, host, realname, raw=arg)

    def received_353(self, (arg,)):
        """Process RPL_NAMREPLY and emit NAMES."""
        # last element is channel
        chan = ("chan", arg.argv[-1])

        dct = {}
        servcapab = self.base.p.core.servcapab

        for nick in arg.text.split(" "):
            if not nick:
                # Huh? No nick? Ignore this.
                continue
            if nick[0] in servcapab.cupres:
                dct[nick[1:]] = nick[0]
            else:
                dct[nick] = ""

        self.emit_signal("NAMES", chan, ("names", dct), raw=arg)

    def received_367(self, (arg,)):
        """Process RPL_BANLIST and emit SETBAN."""
        chan = ("chan", arg.argv[1])
        mask = ("mask", arg.argv[2])
        srcmask = ("srcmask", arg.argv[3])
        time = ("time", int(arg.argv[4]))
        self.emit_signal("SETBAN", chan, mask, srcmask, time, raw=arg)

    def received_391(self, (arg,)):
        """Process RPL_TIME and emit TIME."""
        t = parse_timestamp(arg.text)
        if t is None:
            print "WARNING: unable to convert timestamp"
        else:
            self.emit_signal("SETTIME", ("time", t), raw=arg)

    def received_invite(self, (arg,)):
        """Process INVITE and emit INVITE."""
        nick = ("nick", arg.nick)
        self.emit_signal("INVITE", nick, ("chan", arg.text), raw=arg)

    def received_join(self, (arg,)):
        """Process JOIN and emit JOINs."""
        nick = ("nick", arg.nick)
        for chan in arg.text.split(","):
            self.emit_signal("JOIN", nick, ("chan", chan), raw=arg)

    def received_kick(self, (arg,)):
        """Process KICK and emit KICK."""
        nick = ("nick", arg.nick)
        chan = ("chan", arg.argv[0])
        target = ("target", arg.argv[1])
        reason = ("reason", arg.text)
        self.emit_signal("KICK", nick, chan, target, reason, raw=arg)

    def received_mode(self, (arg,)):
        """Process MODE and emit combinations of SMODE, AMODE, UMODE, CUMODE,
        ADDBAN and UNBAN."""
        nick = ("nick", arg.nick)
        servcapab = self.base.p.core.servcapab
        # make a copy
        args = arg.argv[:]
        # args start with target and modestring
        target = args.pop(0)

        if servcapab.ischannel(target):
            modes = args.pop(0)
            chan = ("chan", target)
            ischannel = True
        else:
            modes = arg.text
            ischannel = False

        sign = None
        for c in modes:
            if c in '+-':
                sign = c
            if ischannel:
                if c in servcapab.smods:
                    mode = ("mode", sign + c)
                    self.emit_signal("SMODE", nick, chan, mode, raw=arg)
                elif c in servcapab.amods:
                    t = None
                    if sign == '+' or c in servcapab.abmods:
                        t = args.pop(0)
                    mode = ("mode", sign + c)
                    arg = ("arg", t)
                    self.emit_signal("AMODE", nick, chan, mode, arg, raw=arg)
                elif c in servcapab.cumods:
                    t = args.pop(0)
                    mode = ("mode", sign + c)
                    target = ("target", t)
                    self.emit_signal("CUMODE", nick, chan, mode,
                                               target, raw=arg)
                elif c == 'b':
                    t = args.pop(0)
                    mask = ("mask", t)
                    if sign == '-':
                        self.emit_signal("UNBAN", nick, chan, mask, raw=arg)
                    else:
                        self.emit_signal("ADDBAN", nick, chan, mask, raw=arg)
            else: # not a channel
                if c in servcapab.umods:
                    self.emit_signal("UMODE", ("mode", sign + c), raw=arg)

    def received_nick(self, (arg,)):
        """Process NICK and emit NICK."""
        nick = ("nick", arg.nick)
        self.emit_signal("NICK", nick, ("new", arg.text), raw=arg)

    def received_notice(self, (arg,)):
        """Process NOTICE and emit NOTICE and CNOTICE or PNOTICE."""
        nick = ("nick", arg.nick)
        target = ("target", arg.argv[0])
        text = ("text", arg.text)

        self.emit_signal("NOTICE", nick, target, text, raw=arg)

        msg = arg.text
        if len(msg) > 2 and msg[0] == '\x01' and msg[-1] == '\x01':
            text = ("text", msg[1:-1])
            self.emit_signal("CTCPREP", nick, target, text, raw=arg)
        elif self.base.p.core.servcapab.ischannel(arg.argv[0]):
            chan = ("chan", arg.argv[0])
            self.emit_signal("CNOTICE", nick, chan, text, raw=arg)
        else:
            self.emit_signal("PNOTICE", nick, target, text, raw=arg)

    def received_part(self, (arg,)):
        """Process PART and emit PARTs."""
        nick = ("nick", arg.nick)
        reason = ("reason", arg.text)

        for chan in arg.argv[0].split(","):
           self.emit_signal("PART", nick, ("chan", chan), reason, raw=arg)

    def received_privmsg(self, (arg,)):
        """Process PRIVMSG and emit PRIVMSG and CPRIVMSG or PPRIVMSG."""
        nick = ("nick", arg.nick)
        target = ("target", arg.argv[0])
        text = ("text", arg.text)

        self.emit_signal("PRIVMSG", nick, target, text, raw=arg)

        msg = arg.text
        if len(msg) > 2 and msg[0] == '\x01' and msg[-1] == '\x01':
            text = ("text", msg[1:-1])
            self.emit_signal("CTCPREQ", nick, target, text, raw=arg)
        elif self.base.p.core.servcapab.ischannel(arg.argv[0]):
            chan = ("chan", arg.argv[0])
            self.emit_signal("CPRIVMSG", nick, chan, text, raw=arg)
        else:
            self.emit_signal("PPRIVMSG", nick, target, text, raw=arg)

    def received_quit(self, (arg,)):
        """Process QUIT and emit QUIT."""
        nick = ("nick", arg.nick)
        self.emit_signal("QUIT", nick, ("reason", arg.text), raw=arg)

    def received_topic(self, (arg,)):
        """Process TOPIC and emit TOPIC."""
        nick = ("nick", arg.nick)
        chan = ("chan", arg.argv[0])
        self.emit_signal("TOPIC", nick, chan, ("topic", arg.text), raw=arg)
