#!/usr/bin/env python2.3
#
#   mibcip - Modular IRC Bot Core In Python
#   Copyright (C) 2003-2004  Helmut Grohne
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import socket
import errno
import select
import time

class _scheduler_event:
    """A scheduler event is a function with a single argument that is to be
    called on some specific point of time. To be able to be canceled it must
    have a reference to the scheduler."""
    def __init__(self, when, func, arg, sched):
        """when is some comparable value for the scheduler sched. On calling
        a _scheduler_event object func(arg) gets called."""
        self.when = when
        self.func = func
        self.arg = arg
        self.sched = sched

    def __call__(self):
        """Actually call the function with the given argument."""
        return self.func(self.arg)

    def __cmp__(self, other):
        """Compare this event to another by comparing the time till execution.
        """
        return cmp(self.when, other.when)

    def cancel(self):
        """Cancel this event by telling the scheduler about it. This function
        is needed to be compatible to twisted."""
        self.sched.cancel(self)

class _scheduler_class:
    """Event scheduling class calls functions after a specific time period.
    This may be used for regular events or to transpose events."""
    def __init__(self):
        self.queue = []

    def add(self, when, func, arg):
        """Add a new event in when seconds. Then func is called with arg as
        single argument."""
        ev = _scheduler_event(when + time.time(), func, arg, self)
        start = 0
        end = len(self.queue)
        while start < end:
            mid=(start+end)/2
            if self.queue[mid] > ev:
                end=mid
            else:
                start=mid+1
        self.queue.insert(start, ev)
        return ev

    def call(self):
        """Call an event."""
        ev = self.queue.pop(0)
        ev()

    def when(self):
        """Check if there is an event waiting for beeing called. Returns -1 for
        no event, 0 for event now or time in seconds to be waited for the
        next event."""
        if bool(self.queue):
            t = self.queue[0].when - time.time()
            if t < 0:
                t = 0
            return t
        else:
            return -1

    def __nonzero__(self):
        return bool(self.queue)

    def cancel(self, ev):
        """Prevent execution of given event."""
        try:
            self.queue.remove(ev)
        except ValueError:
            pass

class _transport_class:
    """This class should behave like twisted's tcp transport. It therefore
    defines read, write and loseConnection."""
    def __init__(self, sock, react):
        self._sock = sock
        self._reactor = react

    def write(self, data):
        """Twisted compatibility."""
        return self._sock.send(data)

    def read(self, l):
        """Twisted compatibility."""
        return self._sock.recv(l)

    def loseConnection(self):
        """Twisted compatibility."""
        fd = self._sock.fileno()
        self._reactor.fdmap[fd].instance.connectionLost(None)
        del self._reactor.fdmap[fd]
        self._sock.close()

class reactorClass:
    """This class should replace the twisted reactor class it defines many of
    the needed methods, but not all of them."""
    def __init__(self):
        self.sched = _scheduler_class()
        self.fdmap = {}

    def callLater(self, time, function):
        """Call a function after a specific amount of time has passed. Behaves
        mostly like the twisted equivalent."""
        return self.sched.add(time, lambda a: function(), ())

    def connectTCP(self, addr, port, factory):
        """Make given factory connect to given address and port. Twisted
        compatibility."""
        factory.addr = (addr,port)
        factory.sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        factory.sock.setblocking(0)

        self.fdmap[factory.sock.fileno()]=factory

        factory.instance = factory.protocol()
        factory.instance.factory = factory
        factory.instance.transport = _transport_class(factory.sock, self)

        err = factory.sock.connect_ex(factory.addr)
        if err in (errno.EINPROGRESS, errno.EALREADY, errno.EWOULDBLOCK):
            factory.established = False
        elif err in (0, errno.EISCONN):
            factory.established = True
        else:
            raise socket.error, err

    def run(self):
        """Let the whole thing run. Twisted compatibility."""
        while self.fdmap:
            r = [] ; w = []
            t = self.sched.when()
            while t == 0:
                self.sched.call()
                t = self.sched.when()
            for (fd,fact) in self.fdmap.items():
                if fact.established:
                    r.append(fd)
                else:
                    w.append(fd)
            try:
                if self.sched:
                    r,w,e = select.select(r,w,[],t)
                else:
                    r,w,e = select.select(r,w,[])
            except select.error, err:
                if err[0] != errno.EINTR:
                    raise err
                continue

            for fd in r:
                inst = self.fdmap[fd].instance
                inst.dataReceived(inst.transport.read(1024))
            for fd in w:
                # a write implies we are now connected.
                fact = self.fdmap[fd]
                fact.established = True
                fact.instance.connectionMade()

class Protocol:
    """This class is empty, but supplies a sufficient interface as described by
    twisted.internet.protocol.Protocol."""

class ClientFactory:
    """This class is empty, but supplies a sufficient interface as described by
    twisted.interface.protocol.ClientFactory."""

reactor=reactorClass()
